DeathClock = CreateFrame("Frame")
DeathClock.version = GetAddOnMetadata("DeathClock", "Version");
DeathClock.logAge = 10
DeathClock.DB = {}
DeathClock.Frames = {}
DeathClock.panel = {}
DeathClock.debug = {}
DeathClock.defaults = {
	["alignment"] = "CENTER",
	["nameplate"] = true,
	["player"] = true,
	["target"] = true,
	["targettarget"] = true,
	["targettargettarget"] = true,
	["focus"] = true,
	["focustarget"] = true,
	["pet"] = true,
	["pettarget"] = true,
	["party"] = false,
	["raid"] = false,
	["boss"] = true,
	["arena"] = true
}
DeathClockSettings = {}

function DeathClock:OnEvent(event, ...)
	-- Event Handler
	self[event](self, event, ...)
end

local function isempty(s)
  return s == nil or s == ''
end

function DeathClock:ADDON_LOADED(event, addonName)
	if (addonName == "DeathClock") then
		--Load Saved Variables
		for option, default in pairs(DeathClock.defaults) do
			if isempty(DeathClockSettings[option]) then
				DeathClockSettings[option] = default
			end
		end
		
		--Create Options Panel
		DeathClock:CreatePanel()
		
		--Create Slash Handler
		SlashCmdList["DeathClock"] = DeathClock.SlashHandler
		SLASH_DEATHCLOCK1 = "/DeathClock"
		SLASH_DEATHCLOCK2 = "/Death"
		SLASH_DEATHCLOCK3 = "/DC"
		SlashCmdList.DEATHCLOCK = function(msg, editBox)
			--InterfaceOptionsFrame_OpenToCategory(DeathClock.panel)
			DeathClock.Bark()
		end
	end
end

function DeathClock:PLAYER_LOGIN()
	if IsAddOnLoaded("ElvUI") then
		DeathClock.CreateOrUpdateFrame("player", "player", ElvUF_Player.Health)
		DeathClock.CreateOrUpdateFrame("pet", "pet", ElvUF_Pet.Health)
		DeathClock.CreateOrUpdateFrame("pettarget", "pettarget", ElvUF_PetTarget.Health)
		DeathClock.CreateOrUpdateFrame("target", "target", ElvUF_Target.Health)
		DeathClock.CreateOrUpdateFrame("targettarget", "targettarget", ElvUF_TargetTarget.Health)
		DeathClock.CreateOrUpdateFrame("targettargettarget", "targettargettarget", ElvUF_TargetTargetTarget.Health)
		DeathClock.CreateOrUpdateFrame("focus", "focus", ElvUF_Focus.Health)
		DeathClock.CreateOrUpdateFrame("focustarget", "focustarget", ElvUF_FocusTarget.Health)
	else
		DeathClock.CreateOrUpdateFrame("player", "player", PlayerFrame.PlayerFrameContent.PlayerFrameContentMain.HealthBarArea.HealthBar, true)
		DeathClock.CreateOrUpdateFrame("pet", "pet", PetFrameHealthBar)
		DeathClock.CreateOrUpdateFrame("target", "target", TargetFrame.TargetFrameContent.TargetFrameContentMain.HealthBar, true)
		DeathClock.CreateOrUpdateFrame("targettarget", "targettarget", TargetFrameToT.HealthBar)
		DeathClock.CreateOrUpdateFrame("focus", "focus", FocusFrame.TargetFrameContent.TargetFrameContentMain.HealthBar, true)
		DeathClock.CreateOrUpdateFrame("focustarget", "focustarget", FocusFrameToT.HealthBar)	
	end
end

-- REFERENCE FROM PLATER
-- local namePlate = C_NamePlate.GetNamePlateForUnit (unitID)
-- local unitFrame = namePlate.unitFrame --unitFrame is the main frame where all things is attached, it has SetAllPoints() on the namePlate frame.
-- local healthBar = unitFrame.healthBar
-- local castBar = unitFrame.castBar

function DeathClock:NAME_PLATE_UNIT_ADDED(event, unitID)
	if (not DeathClockSettings["nameplate"]) then return end

	for i, namePlate in ipairs(C_NamePlate.GetNamePlates()) do
		-- Blizzard = namePlate.UnitFrame.healthBar
		-- Plater = namePlate.unitFrame.healthBar
		-- ElvUI = namePlate.UnitFrame.Health
		--print(namePlate.namePlateUnitToken)
		unitFrame = namePlate.unitFrame or namePlate.UnitFrame
		healthBar = unitFrame.healthBar or unitFrame.Health
		DeathClock.CreateOrUpdateFrame(namePlate.namePlateUnitToken, "nameplate", healthBar, false)
	end
end

function DeathClock:NAME_PLATE_UNIT_REMOVED(event, unitID)
	DeathClock:NAME_PLATE_UNIT_ADDED(event, unitID)
end

function DeathClock.CreateOrUpdateFrame(unitID, unitSetting, attachFrame, hideOnMouseOver)
	if (DeathClockSettings[unitID] == false) then return end
	if (not DeathClock.Frames[unitID]) then
		DeathClock.Frames[unitID] = DeathClockFrame:new(unitID, unitSetting, attachFrame, hideOnMouseOver)
	else
		DeathClock.Frames[unitID]:Update(unitID, attachFrame)
	end
end

function tprint (tbl, indent)
  if not indent then indent = 0 end
  local toprint = string.rep(" ", indent) .. "{\r\n"
  indent = indent + 2 
  for k, v in pairs(tbl) do
    toprint = toprint .. string.rep(" ", indent)
    if (type(k) == "number") then
      toprint = toprint .. "[" .. k .. "] = "
    elseif (type(k) == "string") then
      toprint = toprint  .. k ..  "= "   
    end
    if (type(v) == "number") then
      toprint = toprint .. v .. ",\r\n"
    elseif (type(v) == "string") then
      toprint = toprint .. "\"" .. v .. "\",\r\n"
    elseif (type(v) == "table") then
      toprint = toprint .. tprint(v, indent + 2) .. ",\r\n"
    else
      toprint = toprint .. "\"" .. tostring(v) .. "\",\r\n"
    end
  end
  toprint = toprint .. string.rep(" ", indent-2) .. "}"
  return toprint
end


function DeathClock:COMBAT_LOG_EVENT_UNFILTERED()
	-- Get combat log data
	
	-- event = Event Name
	-- arg1  = Timestamp or Addon Name
	-- arg2  = Damage Type
	-- arg8  = Target GUID
	-- arg12 = Damage Amount (SWING)			extraArg1
	-- arg13 = Damage Amount (ENVIRONMENTAL)	extraArg2
	-- arg15 = Damage Amount (all other types)	extraArg4
	--local timestamp, damageType, hideCaster, sourceGUID, sourceName, sourceFlags, sourceRaidFlags, destGUID, destName, destFlags, destRaidFlags, extraArg1, extraArg2, extraArg3, extraArg4, extraArg5, extraArg6, extraArg7, extraArg8, extraArg9, extraArg10 = CombatLogGetCurrentEventInfo();
	local timestamp, damageType, _, _, _, _, _, destGUID, _, _, _, extraArg1, extraArg2, _, extraArg4 = CombatLogGetCurrentEventInfo()
	if (damageType == "SWING_DAMAGE") then
		DeathClock:logDamage(timestamp, destGUID, extraArg1)
	elseif (damageType == "SPELL_DAMAGE") then
		DeathClock:logDamage(timestamp, destGUID, extraArg4)
	elseif (damageType == "RANGE_DAMAGE") then
		DeathClock:logDamage(timestamp, destGUID, extraArg4)
	elseif (damageType == "SPELL_PERIODIC_DAMAGE") then
		DeathClock:logDamage(timestamp, destGUID, extraArg4)
	elseif (damageType == "SPELL_BUILDING_DAMAGE") then
		DeathClock:logDamage(timestamp, destGUID, extraArg4)
	elseif (damageType == "DAMAGE_SHIELD") then
		DeathClock:logDamage(timestamp, destGUID, extraArg4)
	elseif (damageType == "ENVIRONMENTAL_DAMAGE") then
		DeathClock:logDamage(timestamp, destGUID, extraArg2)
	end
end

function DeathClock:logDamage(timestamp, guID, damage)
	-- Add damage logs to database
	--print(timestamp.." | "..guID.." | "..damage);
	local record = {timestamp, guID, damage};
	tinsert(DeathClock.DB, 1, record);
end

function round(x)
	if x % 1 >= 0.5 then
		return ceil(x)
	else
		return floor(x)
	end
end

function DeathClock:getDamageOnUnit(unitGUID)
	-- tableformat = {time, guid, damage}
	local totalDamage = 0;
	for index, record in ipairs(DeathClock.DB) do
		if (time() - record[1]) > DeathClock.logAge then
			-- Delete Expired Data
			tremove(DeathClock.DB, index);
		elseif record[2] == unitGUID then
			totalDamage = totalDamage + record[3];
		end	
	end
	local dps = round(totalDamage / DeathClock.logAge);
	return dps;
end

function DeathClock:getETD(health, dps)
	local etd;
	if (dps > 0) then
		etd = health / dps;
	else
		etd = -1;
	end
	return etd;
end

function DeathClock:formatTime(seconds)
	local minutes = seconds / 60
	local hours = minutes / 60
	local days = hours / 24
	local years = days / 365
	if (seconds < 1) then
		return "";
	elseif (minutes < 2) then
		return round(seconds) .. "s"
	elseif (hours < 1) then
		return round(minutes) .. "m"
	elseif (days < 1) then
		return round(hours) .. "h"
	elseif (years < 1) then
		return round(days) .. "d"
	elseif (years < 99) then
		return round(years) .. "y"
	else -- Greater than 99 years
		return("∞");
	end
end

function DeathClock:formatDPS(dps)
	local thousand = dps / 1000
	local million = thousand / 1000
	local billion = million / 1000
	local trillion = billion / 1000
	if (dps < 1) then
		return "";
	elseif (thousand < 1) then
		return round(dps)
	elseif (million < 1) then
		return round(thousand) .. "K"
	elseif (billion < 1) then
		return round(million) .. "M"
	elseif (trillion < 1) then
		return round(billion) .. "B"
	else
		return round(trillion) .. "T"
	end
end

function DeathClock.Bark()
	local dps = DeathClock:getDamageOnUnit(UnitGUID("target"))
	local etd = DeathClock:getETD(UnitHealth("target"), dps)
	if etd < 1 then return end
	local msg = "[DeathClock]: <" .. UnitName("target") .. "> will die in "..DeathClock:formatTime(etd).." @ "..DeathClock:formatDPS(dps).." DPS"
	--(msg)
	SendChatMessage(msg, IsInGroup(LE_PARTY_CATEGORY_INSTANCE) and "INSTANCE_CHAT" or IsInRaid() and "RAID" or IsInGroup() and "PARTY" or "SAY");
end

DeathClock:RegisterEvent("ADDON_LOADED")
DeathClock:RegisterEvent("PLAYER_LOGIN")
DeathClock:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
DeathClock:RegisterEvent("NAME_PLATE_UNIT_ADDED")
DeathClock:RegisterEvent("NAME_PLATE_UNIT_REMOVED")
DeathClock:SetScript("OnEvent", DeathClock.OnEvent)

DeathClockFrame = {}
function DeathClockFrame:new(unitID, unitSetting, attachFrame, hideOnMouseOver)
	self = {}
	
	-- Member Variables:
	self.unitID = unitID
	self.unitSetting = unitSetting
	self.attachFrame = attachFrame
	self.hideOnMouseOver = hideOnMouseOver
	self.test = "NA"
	self.lastUpdate = 0
	
	-- Member Functions:
	function self.OnUpdate()
		if time() > self.lastUpdate then
			self:Refresh()
		end
	end
	
	function self.OnEvent(_, event, arg1)
		if (event == "PLAYER_TARGET_CHANGED") then
			self:Refresh()
		elseif (event == "NAME_PLATE_UNIT_ADDED") then
			self:Refresh()
		end
	end
	
	function self.Refresh()
		if (self.hideOnMouseOver and MouseIsOver(self.overlay)) or (DeathClockSettings[self.unitSetting] == false) then
			self.overlay:SetText("")
		else
			--if UnitGUID(self.unitID) == nil then print("nil") end
			dps = DeathClock:getDamageOnUnit(UnitGUID(self.unitID))
			etd = DeathClock:getETD(UnitHealth(self.unitID), dps)
			self.overlay:SetJustifyH(DeathClockSettings["alignment"])
			self.overlay:SetText(DeathClock:formatTime(etd))
			self.lastUpdate = time()
		end
	end
	
	function self.Update(_, unitID, attachFrame)
		--print(unitID)
		self.unitID = unitID
		self.attachFrame = attachFrame
		self.overlay:ClearAllPoints()
		self.overlay:SetPoint("TOPLEFT", self.attachFrame, "TOPLEFT", 0, 0)
		self.overlay:SetPoint("BOTTOMRIGHT", self.attachFrame, "BOTTOMRIGHT", 0, 0)
		
	end
	
	self.frame = CreateFrame("Frame")
	self.frame:RegisterEvent("PLAYER_TARGET_CHANGED")
	self.frame:RegisterEvent("NAME_PLATE_UNIT_ADDED")
	self.frame:SetScript("OnEvent", self.OnEvent)
	self.frame:HookScript("OnUpdate", self.OnUpdate)
	
	self.overlay = self.frame:CreateFontString("DeathClockFont"..self.unitID, "ARTWORK", "TextStatusBarTextLarge")
	self.overlay:SetPoint("TOPLEFT", self.attachFrame, "TOPLEFT", 0, 0)
	self.overlay:SetPoint("BOTTOMRIGHT", self.attachFrame, "BOTTOMRIGHT", 0, 0)
	self.overlay:SetJustifyH(DeathClockSettings["alignment"])
	
	if self.hideOnMouseOver then
		self.attachFrame:HookScript("OnEnter", self.Refresh)
	end
	
	return self
end

function DeathClock:CreatePanel()
	local panel = CreateFrame("Frame");
	panel.name = "DeathClock";
	
	panel.title = panel:CreateFontString(nil, nil, "GameFontNormalLarge");
	panel.title:SetPoint("TOPLEFT", panel, "TOPLEFT", 15, -15);
	panel.title:SetText("DeathClock");
	
	panel.title.small = panel:CreateFontString(nil, nil, "GameFontNormalSmall");
	panel.title.small:SetPoint("BOTTOMLEFT", panel.title, "BOTTOMRIGHT", 5, 0);
	panel.title.small:SetText("version " .. DeathClock.version);
	
	panel.author = panel:CreateFontString(nil, nil, "GameFontNormalSmall");
	panel.author:SetPoint("TOPLEFT", panel.title, "BOTTOMLEFT", 10, 0);
	panel.author:SetText("by Ne0nguy (Portman-BleedingHollow)");
	
	panel.description = panel:CreateFontString(nil, nil, "GameFontHighlight");
	panel.description:SetPoint("TOPLEFT", panel.title, "BOTTOMLEFT", 0, -20);
	panel.description:SetJustifyH("LEFT")
	panel.description:SetText("DeathClock calculates the the estimated time until death\nbased on the average damage per second over the last 10 seconds.")
	
	panel.instruction = panel:CreateFontString(nil, nil, "GameFontNormal");
	panel.instruction:SetPoint("TOPLEFT", panel.description, "BOTTOMLEFT", 0, -10);
	panel.instruction:SetJustifyH("LEFT")
	panel.instruction:SetText("Select where you would like DeathClock to appear:")

	panel.checkNamePlates = DeathClock:CreateOption("nameplate", "Name Plates", panel, panel.instruction, -5)
	
	panel.menuAlign = CreateFrame("FRAME", nil, panel, "UIDropDownMenuTemplate")
	panel.menuAlign:SetPoint("LEFT", panel.checkNamePlates.Text, "RIGHT", 0, 0);
	UIDropDownMenu_SetWidth(panel.menuAlign, 150)
	UIDropDownMenu_SetText(panel.menuAlign, "Alignment: "..DeathClockSettings["alignment"])
	
	local menuList = {"LEFT", "CENTER", "RIGHT"}
	UIDropDownMenu_Initialize(panel.menuAlign, function()
		local info = UIDropDownMenu_CreateInfo()
		for i, v in pairs(menuList) do
			info.text = v
			info.checked = (DeathClockSettings["alignment"] == v)
			info.func = function(self)
				DeathClockSettings["alignment"] = self.value
				UIDropDownMenu_SetText(panel.menuAlign, "Alignment: "..DeathClockSettings["alignment"])
			end
			UIDropDownMenu_AddButton(info)
		end
	end)

	-- Implement the function to change the favoriteNumber
	function panel.menuAlign:SetValue(newValue)
		favoriteNumber = newValue
		-- Update the text; if we merely wanted it to display newValue, we would not need to do this
		UIDropDownMenu_SetText(panel.menuAlign, "Favorite number: " .. favoriteNumber)
		-- Because this is called from a sub-menu, only that menu level is closed by default.
		-- Close the entire menu with this next call
		CloseDropDownMenus()
	end

	panel.checkPlayer = DeathClock:CreateOption("player", "Player", panel, panel.checkNamePlates, -5)
	
	panel.checkTarget = DeathClock:CreateOption("target", "Target", panel, panel.checkPlayer, -5)
	panel.checkTargetTarget = DeathClock:CreateOption("targettarget", "TargetTarget", panel, panel.checkTarget)
	panel.checkTargetTargetTarget = DeathClock:CreateOption("targettargettarget", "TargetTargetTarget", panel, panel.checkTargetTarget)
	
	panel.checkFocus = DeathClock:CreateOption("focus", "Focus", panel, panel.checkTargetTargetTarget, -5)
	panel.checkFocusTarget = DeathClock:CreateOption("focustarget", "FocusTarget", panel, panel.checkFocus)
	
	panel.checkPet = DeathClock:CreateOption("pet", "Pet", panel, panel.checkFocusTarget, -5)
	panel.checkPetTarget = DeathClock:CreateOption("pettarget", "PetTarget", panel, panel.checkPet)
	
	panel.checkParty = DeathClock:CreateOption("party", "Party (not implemented)", panel, panel.checkPetTarget, -5)
	panel.checkRaid = DeathClock:CreateOption("raid", "Raid (not implemented)", panel, panel.checkParty)
	panel.checkBoss = DeathClock:CreateOption("boss", "Boss (not implemented)", panel, panel.checkRaid)
	panel.checkArena = DeathClock:CreateOption("arena", "Arena (not implemented)", panel, panel.checkBoss)
	
	panel.tip = panel:CreateFontString(nil, nil, "GameFontNormalLarge");
	panel.tip:SetPoint("TOPLEFT", panel.checkArena, "BOTTOMLEFT", 0, -10);
	panel.tip:SetJustifyH("LEFT")
	panel.tip:SetText("Tip: ")
	
	panel.tip.text = panel:CreateFontString(nil, nil, "GameFontHighlight");
	panel.tip.text:SetPoint("LEFT", panel.tip, "RIGHT");
	panel.tip.text:SetJustifyH("LEFT")
	panel.tip.text:SetText("Typing /DeathClock /Death or /DC will post the DeathClock for the current target in chat.\nMake a /macro containing one of those slash commands to add a button to your action bar.")
	
	DeathClock.panel = panel
	InterfaceOptions_AddCategory(DeathClock.panel)
end

function DeathClock:CreateOption(unit, name, parent, attach, offset)
	local btn = CreateFrame("CheckButton", nil, parent, "InterfaceOptionsCheckButtonTemplate")
	btn:SetPoint("TOPLEFT", attach, "BOTTOMLEFT", 0, offset or 0)
	btn:SetChecked(DeathClockSettings[unit])
	btn:HookScript("OnClick", function(_, self, down)
		DeathClockSettings[unit] = btn:GetChecked()
	end)
	btn.Text:SetPoint("LEFT", btn, "RIGHT", 5, 0)
	btn.Text:SetText(name)
	
	return btn
end